private import SDL;

class FpsTimer {
public:
	this(bit enable) {
		fps_ = 60;
		ticksPerFrame_ = 1000 / 60;
		shouldSkip_ = false;
		turns_ = 0;
		systemDropTurns_ = 0;
		frameDropTurns_ = 0;
		prevDrawTicks_ = 0;
		enable_ = enable;
		slow_ = 0;
        frameSkip_ = true;
	}

    void setFrameSkip(bool f) { frameSkip_ = f; }

	void setFps(int fps) {
		fps_ = fps;
		ticksPerFrame_ = 1000 / fps;
	}

	void start() {
		startTicks_ = ticks_ = SDL_GetTicks();
		turns_ = frameDropTurns_ = 0;
	}

	void slow(int s) {
		slow_ = s;
	}

	void wait() {
		ticks_ += ticksPerFrame_;
		if (slow_) ticks_ += ticksPerFrame_ >> slow_;

		shouldSkip_ = false;
		turns_++;

        if (!frameSkip_) return;

		int now = SDL_GetTicks();

		if (ticks_ < now) {
//			if (prevDrawTicks_ > now - 100) {
				frameDropTurns_++;
				shouldSkip_ = true;
/*
			}
			else {
				systemDropTurns_++;
				prevDrawTicks_ = now;
				ticks_ = now;
			}
*/
			return;
		}

		prevDrawTicks_ = now;
		if (enable_) SDL_Delay(ticks_ - now);
	}

	int getTrueFps() {
		if (ticks_ - startTicks_ != 0) {
			return (turns_ - frameDropTurns_) * 1000 /
				(SDL_GetTicks() - startTicks_);
		}
		else {
			return 0;
		}
	}

	bit shouldSkip() { return shouldSkip_; }

	int getTurn() { return turns_; }
	int getSystemDroppedTurn() { return systemDropTurns_; }
	int getDroppedTurn() { return frameDropTurns_; }
	int getFps() { return fps_; }
	int getEllapsedTicks() { return ticks_ - startTicks_; }
	int getEllapsedSecond() { return getEllapsedTicks() / 1000; }

	int getDropPercent() {
		if (turns_ == 0) return 0;
		return 100 * frameDropTurns_ / turns_;
	}

private:
	int startTicks_;
	int fps_;
	int ticksPerFrame_;
	int ticks_;
	bit shouldSkip_;
	int turns_;
	int frameDropTurns_;
	int systemDropTurns_;
    int prevDrawTicks_;
	bit enable_;
	int slow_;
    bool frameSkip_;

};
