import log;

import std.c.time;
import std.string;
import std.socket;
import std.socketstream;

version (linux) {
    const int SOL_SOCKET = 1;
}
version (Win32) {
    const int SOL_SOCKET = 0xffff;
}
version (darwin) {
    const int SOL_SOCKET = 0xffff;
}

version (THREAD) {
    import std.thread;

    import SDL_thread;

    class Thread {
        extern (C) static int invoke_thread(void* self) {
            return (cast(Thread)self).run();
        }

        void start() {
            th = SDL_CreateThread(&invoke_thread, cast(void*)this);
        }

        int run() {
            assert(false);
        }

        void wait() {
            SDL_WaitThread(th, null);
        }

        SDL_Thread* th;
    }

    class Srv : Thread {
        char[][] ops;
        bool done;
        string[] events;

        char[][] retrieveOps() {
            synchronized(this) {
                char[][] r;
                foreach (o; ops) {
                    r ~= o.dup;
                }
                ops.length = 0;
                return r;
            }
        }

        void addEvent(string ev) {
            synchronized(this) {
                events ~= ev;
            }
        }

        void finish() {
            Socket sock = new Socket(AddressFamily.INET,
                                     SocketType.STREAM,
                                     ProtocolType.IP);
            sock.connect(new InternetAddress(2525));
            sock.send(`done`);
            sock.close();
            wait();
        }

        void proc() {
        }

        override int run() {
            /+
             signal(30 /* SIGUSR1 */, 1 /* SIG_IGN */);
             signal(31 /* SIGUSR2 */, 1 /* SIG_IGN */);
             +/

            InternetAddress serverIp = new InternetAddress(2525);
            Socket serverSocket = new Socket(AddressFamily.INET,
                                             SocketType.STREAM,
                                             ProtocolType.IP);
            int one = 1;
            serverSocket.setOption(
                cast(SocketOptionLevel)(SocketOptionLevel.SOL_SOCKET),
                SocketOption.REUSEADDR, one);
            serverSocket.bind(serverIp);
            serverSocket.listen(1);

            Log.info("server thread started");

            while (true) {
                Socket sock = serverSocket.accept();

                Address ip = sock.remoteAddress();
                Log.info("connection established: " ~ ip.toString());

                SocketStream ss = new SocketStream(sock);
                foreach (char[] line; ss) {
                    line = chomp(line);
                    if (line == `done`) {
                        done = true;
                        break;
                    }
                    else if (line == `poll`) {
                        Log.info(`poll: ` ~ line);
                        while (true) {
                            synchronized (this) {
                                if (events.length) {
                                    string ev = events[0];
                                    events = events[1..$];
                                    Log.info(`event: ` ~ ev);
                                    ss.writeLine(ev);
                                    break;
                                }
                            }
                            usleep(100000);
                        }
                        continue;
                    }
                    synchronized(this) {
                        ops ~= line;
                    }
                }

                Log.info("connection closed: " ~ ip.toString());
                sock.close();
            }

            return 0;
        }
    }
}
else {
    class Srv {
        char[][] ops;
        bool done;
        string[] events;

        Socket serverSocket;
        Socket sock;
        Address ip;
        SocketStream ss;
        bool polling;

        char[][] retrieveOps() {
            synchronized(this) {
                char[][] r;
                foreach (o; ops) {
                    r ~= o.dup;
                }
                ops.length = 0;
                return r;
            }
        }

        void addEvent(string ev) {
            synchronized(this) {
                events ~= ev;
            }
        }

        void finish() {
            if (sock) {
                sock.close();
                sock = null;
            }
            serverSocket.close();
        }

        void start() {
            InternetAddress serverIp = new InternetAddress(2525);
            serverSocket = new Socket(AddressFamily.INET,
                                      SocketType.STREAM,
                                      ProtocolType.IP);
            int one = 1;
            serverSocket.setOption(cast(SocketOptionLevel)(SOL_SOCKET),
                SocketOption.REUSEADDR, one);
            serverSocket.bind(serverIp);
            serverSocket.listen(1);
        }

        void proc() {
            if (polling) {
                if (events.length) {
                    string ev = events[0];
                    events = events[1..$];
                    Log.info(`event: ` ~ ev);
                    ss.writeLine(ev);
                    polling = false;
                }
            }
            if (sock) {
                SocketSet sset = new SocketSet(1);
                sset.add(sock);
                if (Socket.select(sset, null, null, 1)) {
                    char[] line = ss.readLine();
                    Log.info(`received: ` ~ line);

                    if (line == `done`) {
                        done = true;
                    }
                    else if (line == `poll`) {
                        Log.info(`poll: ` ~ line);
                        polling = true;
                    }
                    else {
                        ops ~= line;
                    }
                }
                if (ss && ss.eof()) {
                    Log.info("connection closed: " ~ ip.toString());
                    sock.close();
                    sock = null;
                    ss = null;
                    polling = false;
                }
            }
            else {
                SocketSet sset = new SocketSet(1);
                sset.add(serverSocket);
                if (Socket.select(sset, null, null, 1)) {
                    sock = serverSocket.accept();

                    ip = sock.remoteAddress();
                    Log.info("connection established: " ~ ip.toString());
                    ss = new SocketStream(sock);
                }
            }
        }
    }
}
